"""fleks.util.click"""

from click.core import Context as ClickContext


def click_recursive_help(
    cmd,
    parent=None,
    out={},
    dotpath: str = None,
    module: str = None,
    path: str = None,
    # file=sys.stdout
):
    """ """
    # source: adapted from https://stackoverflow.com/questions/57810659/automatically-generate-all-help-documentation-for-click-commands
    full_name = cmd.name
    pname = getattr(cmd, "parent", None)
    pname = parent and getattr(parent, "name", "") or ""
    ctx = ClickContext(cmd, info_name=cmd.name, parent=parent)
    help_txt = cmd.get_help(ctx)
    invocation_sample = help_txt.split("\n")[0]
    for x in "Usage: [OPTIONS] COMMAND [COMMAND] [ARGS] ...".split():
        invocation_sample = invocation_sample.replace(x, "")
    out = {
        **out,
        **{
            full_name: dict(
                name=cmd.name,
                path=path,
                dotpath=dotpath,
                module=module,
                invocation_sample=invocation_sample.lstrip().rstrip(),
                help=help_txt,
            )
        },
    }
    commands = getattr(cmd, "commands", {})
    for sub in commands.values():
        out = {
            **out,
            **click_recursive_help(
                sub, parent=ctx, dotpath=dotpath, module=module, path=path
            ),
        }
    return out



"""fleks.util.types

This module collects common imports and annotation-types, i.e.
various optional/composite types used in type-hints, underneath
one convenient namespace.
"""

import typing
from pathlib import Path as BasePath

from types import *  # noqa
from typing import *  # noqa

from pydantic import BaseModel, Extra, Field, validate_arguments  # noqa

validate = validate_arguments

OptionalAny = typing.Optional[typing.Any]
PathType = type(BasePath())

Bool = bool
NoneType = type(None)

BoolMaybe = typing.Optional[bool]
StringMaybe = typing.Optional[str]
CallableMaybe = typing.Optional[typing.Callable]
DictMaybe = typing.Optional[typing.Dict]
TagDict = typing.Dict[str, str]


Namespace = typing.Dict[str, typing.Any]
CallableNamespace = typing.Dict[str, typing.Callable]


def new_in_class(name: str, kls: typing.Type) -> bool:
    """ """
    return name in dir(kls) and not any([name in dir(base) for base in kls.__bases__])


def is_subclass(x, y, strict=True) -> bool:
    """
    returns True if first argument is a subclass of second one.
    """
    if isinstance(x, (typing.Type)) and issubclass(x, y):
        if strict and x == y:
            return False
        return True
    return False


class classproperty:
    """ """

    def __init__(self, fxn):
        self.fxn = fxn

    def __get__(self, obj, owner) -> OptionalAny:  # noqa
        # assert obj
        return self.fxn(owner)


class classproperty_cached(classproperty):
    """ """

    CLASSPROP_CACHES = {}

    def __get__(self, obj, owner) -> OptionalAny:  # noqa
        # assert obj
        result = self.__class__.CLASSPROP_CACHES.get(self.fxn, self.fxn(owner))
        self.__class__.CLASSPROP_CACHES[self.fxn] = result
        return self.__class__.CLASSPROP_CACHES[self.fxn]



"""fleks.util.console"""

import os


def is_notebook() -> bool:
    """
    Returns true if this runtime is inside a JupyterLab notebook.
    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


def color_disabled() -> bool:
    """see https://no-color.org/"""
    return True if os.environ.get("NO_COLOR", "") else False



""" """

import collections
from gettext import gettext as _

import click


class RootGroup(click.Group):
    """ """

    def __init__(self, *args, **kwargs):
        """ """
        # NB: this is needed, otherwise main help is messy
        kwargs.update(help="")
        super().__init__(*args, **kwargs)

    def format_commands(
        self, ctx: click.Context, formatter: click.HelpFormatter
    ) -> None:
        """"""
        commands = []
        for subcommand in self.list_commands(ctx):
            cmd = self.get_command(ctx, subcommand)
            if cmd is None or cmd.hidden:
                continue
            commands.append((subcommand, cmd))
        commands = dict(commands)
        if len(commands):
            # allow for 3 times the default spacing
            limit = formatter.width - 6 - max(len(cmd[0]) for cmd in commands)
            plugin_subs = {
                name: obj
                for name, obj in commands.items()
                if obj.__class__.__name__ == "Group"
            }
            toplevel = dict(
                core=[],
                meta=collections.defaultdict(collections.defaultdict),
                plugins=collections.defaultdict(list),
            )
            for subcommand, cmd in commands.items():
                help = cmd.get_short_help_str(limit)
                is_plugin = cmd in list(plugin_subs.values())
                label = ""
                if is_plugin:
                    subcom_obj = plugin_subs[subcommand]
                    plugin_class = getattr(subcom_obj, "plugin_class", None)
                    # if issubclass(plugin_kls, (fleks.Plugin,)):
                    cli_label = getattr(plugin_class, "cli_label", None)
                    cli_description = getattr(
                        plugin_class,
                        "cli_description",
                        getattr(
                            plugin_class.__class__,
                            "cli_description",
                            plugin_class.__doc__,
                        ),
                    )
                    if cli_label:
                        toplevel["plugins"][cli_label].append(
                            (f"{subcommand}:", f"{cmd.help}")
                        )
                        toplevel["meta"][cli_label]["description"] = cli_description
                        continue
                toplevel["core"].append((f"{subcommand}:", f"{cmd.help}"))

            if toplevel["core"]:
                order = ["plan", "apply", "config", "config-raw"]
                ordering = []
                for o in order:
                    for subc, subh in toplevel["core"]:
                        if subc == o:
                            ordering.append((subc, subh))
                            toplevel["core"].remove((subc, subh))
                toplevel["core"] = ordering + toplevel["core"]
                with formatter.section(_(click.style("Top-level", bold=True))):
                    formatter.write_text(
                        click.style(
                            "Core functionality (these names are forbid to plugins)",
                            dim=True,
                        )
                    )
                    formatter.write_dl(toplevel["core"])
            for label in toplevel["plugins"]:
                cli_description = toplevel["meta"][label]["description"]
                with formatter.section(_(click.style(f"{label.title()}", bold=True))):
                    if cli_description:
                        formatter.write_text(
                            click.style(f"{cli_description.lstrip()}", dim=True)
                        )
                    formatter.write_dl(toplevel["plugins"][label])

    def format_usage(self, ctx, formatter):
        """
        :param ctx: param formatter:
        """
        # terminal_width, _ = click.get_terminal_size()
        terminal_width = 30
        click.echo("-" * terminal_width)
        super().format_usage(ctx, formatter)

    def parse_args(self, ctx, args):
        default = self.default
        originals = [args.copy(), ctx.__dict__.copy()]
        copy = [x for x in args.copy() if x != "--help"]
        ctx2 = default.make_context("default", copy)
        with ctx2:
            default.invoke(ctx2)
        return super(click.Group, self).parse_args(ctx, args)



"""fleks.__main__"""

from pathlib import Path

from fleks import cli
from fleks.util import lme

LOGGER = lme.get_logger(__name__)
DEFAULT_INPUT_FILE = "/dev/stdin"


@cli.click.group(name=Path(__file__).parents[0].name)
def entry():
    """CLI tool for `fleks` library"""


if __name__ == "__main__":
    entry()



"""fleks.config"""

import pydash

from fleks import models
from fleks.util import typing


class ConfProto:
    def __getitem__(
        self,
        key,
    ):
        if isinstance(key, (slice,)):
            start, stop, step = key.start, key.stop, key.step
            raise Exception("niy")
        tmp = self.dict(
            # exclude_unset=True,
            by_alias=True
        )
        try:
            return tmp[key]
        except (KeyError, TypeError) as exc:
            return pydash.get(tmp, key)

    def __repr__(self):
        return f"<{self.__class__.__name__}['{self.__class__.config_key}']>"


from pydantic import ConfigDict


class Config(models.BaseModel, ConfProto):
    """ """

    config_key: typing.ClassVar[str] = typing.Field(default=None)
    # debug = False
    # parent = None
    # priority = 100
    # override_from_base = True

    def __init__(self, **this_config) -> None:
        """ """
        kls_defaults = getattr(self.__class__, "defaults", {})
        super().__init__(**{**kls_defaults, **this_config})
        # self._bootstrappery=this_config
        # conflicts = []
        # for pname in self.__class__.__properties__:
        #     if pname in called_defaults or pname in kls_defaults:
        #         conflicts.append(pname)
        #         continue
        #     else:
        #         self[pname] = getattr(self, pname)
        # self.resolve_conflicts(conflicts)


class StrictConfig(models.StrictBaseModel, ConfProto):
    model_config = ConfigDict(strict=True)

    # def as_dict(self, **kwargs):
    #     kwargs.update(
    #         # exclude_unset=True,
    #         by_alias=True
    #     )
    #     return self.dict(**kwargs)

    # @typing.classproperty
    # def logger(kls):
    #     """
    #
    #     :param kls:
    #
    #     """
    #     return lme.get_logger(f"{kls._logging_name}")

    # def resolve_conflicts(self, conflicts: typing.List) -> None:
    #     """
    #
    #     :param conflicts: typing.List:
    #     :param conflicts: typing.List:
    #
    #     """
    #     conflicts and LOGGER.info(
    #         f"'{self.config_key}' is resolving {len(conflicts)} conflicts.."
    #     )
    #
    #     record = collections.defaultdict(list)
    #     for pname in conflicts:
    #         prop = getattr(self.__class__, pname)
    #         strategy = tags.get(prop, {}).get("conflict_strategy", "user_wins")
    #         if strategy == "user_wins":
    #             record[strategy].append(f"{self.config_key}.{pname}")
    #         elif strategy == "override":
    #             val = getattr(self, pname)
    #             self[pname] = val
    #             record[strategy] += [pname]
    #         else:
    #             LOGGER.critical(f"unsupported conflict-strategy! {strategy}")
    #             raise SystemExit(1)
    #
    #     overrides = record["override"]
    #     if overrides:
    #         pass
    #
    #     user_wins = record["user_wins"]
    #     if user_wins:
    #         msg = "these keys have defaults, but user-provided config wins: "
    #         tmp = text.to_json(user_wins)
    #         self.logger.info(f"{msg}\n  {tmp}")
    # def __iter__(self):
    #     return iter(self.as_dict())

    # def __rich__(self):
    #     return self.as_dict()



"""fleks.cli.options:

Common options for CLI
"""

import os

from fleks.cli import click
from fleks.util import lme

LOGGER = lme.get_logger(__name__)

strict = click.flag(
    "--strict",
    help=("if true, runs in strict mode"),
)
script = click.option("--script", default=None, help=("script to use"))
file = click.option("--file", "-f", default="", help=("file to read as input"))
stdout = click.flag("--stdout", help=("whether to write to stdout."))
name = click.option("--name", default="", help=("name to use"))
output = click.option(
    "--output",
    "-o",
    help="when set, output will be written to this file",
    default="/dev/stdout",
)
output = click.option(
    "--output", "-o", default="", help=("output file to write.  (optional)")
)
output_file = click.option(
    "--output",
    "-o",
    metavar="output_file",
    default="",
    help=("output file to write.  (optional)"),
)
output_dir = click.option(
    "--output-dir", default="docs/cli", help=("output directory (optional)")
)
plan = click.flag(
    "--plan",
    "-p",
    "should_plan",
    help="plan only; no action",
)
ignore_missing = click.flag(
    "--ignore-missing",
    help="ignore missing docstrings (only updates empty or out-dated ones)",
)
ignore_private = click.flag(
    "--ignore-private",
    help='ignore names that start with "_")',
)

org_name = click.option(
    "--org-name", "-o", default="", help="defaults to {github.org_name}"
)

inplace = in_place = click.flag(
    "--in-place",
    help=("if true, writes to {file}.{ext} (dropping any other extensions)"),
)

should_print = click.flag(
    "--print",
    "should_print",
    help="if set, displays result on stdout even when `--output <file>` is passed",
)

includes = click.option(
    "--include",
    "includes",
    default=[],
    help=("path to use for template-root / includes"),
    multiple=True,
    # type=list,
    # cls=click.OptionEatAll,
)

ctx = click.option("--ctx", default="", help=("context to use"))
header = click.option(
    "--header", default="", help=("header to prepend output with. (optional)")
)
stdout = click.flag("--stdout", help=("whether to write to stdout."))
format = format_json = click.option(
    "--format", "-m", default="json", help=("output format to write")
)
format_markdown = click.option(
    "--format", "-m", default="markdown", help=("output format to write")
)
format = click.option("--format", "-m", default="json", help=("output format to write"))
package = click.option("--package", "-p", default=os.environ.get("PY_PKG", ""))
file_setupcfg = click.option(
    "--file", "-f", default="setup.cfg", help=("file to grab entrypoints from")
)
module = click.option(
    "--module",
    "-m",
    default="",
    help=("module to grab click-cli from. " "(must be used with `name`)"),
)



"""fleks.util.tagging:

helpers for tagging things with decorators.  this stuff is simple but
pretty abstract; to understand this it's probably better to just look
at how it's used.  you might want to start by just checking out the
tests[1] or the use-cases[2,3] instead.

[1]: tests/units/test_util_tagging.py
[2]: fleks.tagging
"""

from __future__ import annotations

from collections import defaultdict

from fleks.util import lme, typing

LOGGER = lme.get_logger(__name__)


# FIXME: avoid collisions by namespacing like this:
# https://multiple-dispatch.readthedocs.io/en/latest/design.html#namespaces-and-dispatch
GLOBAL_TAG_REGISTRY = defaultdict(dict)


def tag_factory(*args) -> typing.Any:
    """
    :param *args:
    """

    class tagger(dict):
        def get_tags(self, obj: typing.Any) -> dict:
            return GLOBAL_TAG_REGISTRY[obj]

        def __getitem__(self, obj: typing.Any) -> dict:
            return self.get_tags(obj)

    return tagger()


TagDict = typing.Dict[str, typing.Any]


class tagsM:  # FIXME: use data-class
    def __call__(self, **tags: TagDict):
        def decorator(func: typing.Callable) -> typing.Callable:
            merged = {**GLOBAL_TAG_REGISTRY.get(func, {}), **tags}
            # LOGGER.debug(f"tagging {func} with {merged}")
            GLOBAL_TAG_REGISTRY[func] = merged
            return func

        return decorator

    def __getattr__(self, name: str) -> typing.Any:
        if name in "get".split():
            return getattr(GLOBAL_TAG_REGISTRY, name)
        return self[name]

    @typing.validate_arguments
    def __setitem__(self, item: typing.Any, tags: TagDict):
        assert tags is not None
        GLOBAL_TAG_REGISTRY[item] = tags

    # @typing.validate_arguments
    # -> dict[str, typing.Any]
    def __getitem__(self, item: typing.Any) -> TagDict:
        tmp = GLOBAL_TAG_REGISTRY.get(item, {})
        if not tmp and callable(item) and type(item) == typing.MethodType:
            fxn = item
            cfxn = getattr(fxn.__self__.__class__, fxn.__name__)
            tmp = GLOBAL_TAG_REGISTRY.get(cfxn, {})
        tmp = tmp or tag_factory(item)
        self.__setitem__(item, tmp)
        return tmp or {}

    __iter__ = GLOBAL_TAG_REGISTRY.__iter__


tags = tagsM()


def tagged_property(**ftags):
    """Equivalent to:
    @tagging.tags(foo=bar)
    @property
    def method(self):
        ...
    """

    def dec(fxn):
        @tags(**ftags)
        @property
        def newf(*args, **kwargs):
            return fxn(*args, **kwargs)

        return newf

    return dec



"""fleks"""

from memoized_property import memoized_property  # noqa

from .config import Config  # noqa
from .plugin import Plugin  # noqa
from .util import tagging, typing  # noqa
from .util.typing import classproperty  # noqa



"""fleks.models"""

import json
import pathlib

from fleks.util import lme, typing

LOGGER = lme.get_logger(__name__)


class JSONEncoder(json.JSONEncoder):
    """ """

    encoders = {}

    @classmethod
    def register_encoder(kls, type=None, fxn=None):
        """ """
        kls.encoders[type] = fxn

    def encode(self, obj):
        """ """
        result = None
        if callable(getattr(obj, "json", None)):
            return obj.json()
        for _type, fxn in self.encoders.items():
            if isinstance(obj, (_type,)):
                LOGGER.warning(f"{obj} matches {_type}, using {fxn}")
                return fxn(obj)
        return super().encode(obj)

    def default(self, obj):
        """ """
        # FIXME: use multimethod
        if callable(getattr(obj, "dict", None)):
            return obj.dict()
        if callable(getattr(obj, "as_dict", None)):
            return obj.as_dict()
        else:
            enc = self.encoders.get(type(obj), str)
            return enc(obj)


def to_json(obj, cls=None, minified=False, indent: int = 2, **kwargs) -> str:
    """
    custom version of `json.dumps` to always use custom JSONEncoder
    """
    indent = None if minified else indent
    cls = cls if cls is not None else JSONEncoder
    return json.dumps(obj, indent=indent, cls=cls, **kwargs)


JSONEncoder.register_encoder(type=map, fxn=list)
JSONEncoder.register_encoder(type=pathlib.Path, fxn=lambda v: str(v))
JSONEncoder.register_encoder(type=pathlib.PosixPath, fxn=lambda v: str(v))


class MProto(typing.BaseModel):

    def json(self, **kwargs):
        """Overrides pydantic for better serialization"""
        return to_json(self.dict(**kwargs))

    def items(self):
        """dictionary compatability"""
        return self.dict().items()

    @classmethod
    def get_properties(cls):
        return [
            prop
            for prop in dir(cls)
            if isinstance(getattr(cls, prop), property)
            and prop not in ("__values__", "fields")
        ]

    def _dict(
        self,
        *,
        include=None,
        exclude=None,
        by_alias: bool = True,
        skip_defaults: bool = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ):
        """ """
        # -> "DictStrAny":
        # Include and exclude properties
        include = include or set()
        include = include.union(getattr(self.Config, "include", set()))
        if len(include) == 0:
            include = None

        exclude = exclude or set()
        exclude = exclude.union(getattr(self.Config, "exclude", set()))
        if len(exclude) == 0:
            exclude = None
        attribs = super().dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )
        props = self.get_properties()

        if include:
            props = [prop for prop in props if prop in include]
        if exclude:
            props = [prop for prop in props if prop not in exclude]

        # Update the attribute dict with the properties
        if props:
            attribs.update({prop: getattr(self, prop) for prop in props})
        for key, val in attribs.items():
            if isinstance(val, (BaseModel,)):
                attribs[key] = val.dict(
                    include=include,
                    exclude=exclude,
                    by_alias=by_alias,
                    skip_defaults=skip_defaults,
                    exclude_unset=exclude_unset,
                    exclude_defaults=exclude_defaults,
                    exclude_none=exclude_none,
                )
        return attribs

    def dict(self, *args, **kwargs):
        """Overrides pydantic for better serialization"""
        return self._dict(*args, **kwargs)

    def __repr__(self):
        return f"<{self.__class__.__name__}[..]>"

    __str__ = __repr__


class BaseModel(MProto):
    """
    Extends pydantic's BaseModel with better support for
    things like serialization & dynamic values from properties
    """

    class Config:
        extra = typing.Extra.allow
        arbitrary_types_allowed = True
        # https://github.com/pydantic/pydantic/discussions/5159
        frozen = True
        include: typing.Set[str] = set()
        exclude: typing.Set[str] = set()


class StrictBaseModel(MProto):
    class Config:
        extra = typing.Extra.forbid
        arbitrary_types_allowed = False
        # https://github.com/pydantic/pydantic/discussions/5159
        frozen = True
        include: typing.Set[str] = set()
        exclude: typing.Set[str] = set()
